
#include <xc.h>
#include "main.h"
#include "uart.h"


const char HEXITS[16] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};

extern unsigned char UART_RX_buffer[UART_RX_BUFFER_SIZE];       // UART receive buffer
extern unsigned char UART_RX_buffer_size;
extern volatile unsigned char U1RX_flag;                        // new data in the RX buffer


volatile unsigned char UART_TX_buffer[UART_TX_BUFFER_SIZE];     // UART transmit buffer
volatile unsigned int tx_buffer_pos=0;    // number of bytes already sent from the TX buffer
volatile unsigned int tx_buffer_len=0;    // number of valid characters in the TX buffer
volatile unsigned int tx_fifo=0;          // number of bytes in the UART's hardware buffer (hardware glitch prevents us from reading this)
volatile unsigned char U1TX_flag=0;       // old flag that can be used if we don't want a TX buffer


// initialize the UART module
void init_uart()
{
    U1BRG = 12;                 // 9600 baud @ Fcy = 2MHz: U1BRG = 2e6/(16*9600)-1 = 12
    //U1BRG = 32;                 // 9600 baud @ Fcy = 5MHz: U1BRG = 5e6/(16*9600)-1 = 32
    //U1BRG = 127;                 // 9600 baud @ Fcy = 20MHz: U1BRG = 20e6/(16*9600)-1 = 129
    //U1BRG = 21;
    //U1MODEbits.BRGH = 1;
    U1MODEbits.UARTEN = 1;      // enable UART
    U1STAbits.UTXISEL0 = 1;     // interrupt when last character is shifted out of TSR
    IEC0bits.U1TXIE = 1;        // enable TX interrupt
    IFS0bits.U1TXIF = 0;        // clear TX interupt flag
    U1STAbits.UTXEN = 1;        // enable transmit

    U1STAbits.URXISEL = 0b00;   // RX interrupt on any character
    U1MODEbits.RXINV = 0;

    IEC0bits.U1RXIE = 1;        // enable RX interrupt
    IFS0bits.U1RXIF = 0;        // clear RX interupt flag
}

// send a single character
void txChar(unsigned char c)
{
    //while ( U1STAbits.UTXBF || U1STAbits.TRMT == 0);   // wait while Tx buffer full
    //U1TXREG = c;

    // normally we would use UTXBF to determine if the buffer is full, but there is a silicon error on the 24F16KA301 family
    // so we have to manually keep track of the TX FIFO (should be able to put 4 bytes in, but we can only put in 3)
    if (tx_fifo >= 2 || (tx_buffer_len > 0))// transmit buffer is full or we already have buffer characters
    {
        UART_TX_buffer[tx_buffer_len++] = c; // add this character to our buffer
    } else {    // the transmit FIFO has openings, so we can send it there
        U1TXREG = c;
        tx_fifo++;
    }
    // see _U1TXInterrupt for the other part of this code
}

// transmit a string of characters (must be null terminated)
void txStr(char *s)
{
    while( *s)			// loop until *s == '\0' the  end of the string
        txChar( *s++);	// send the character and point to the next one
}

// send a char as two hexits
void txHex(unsigned char in)
{
    txChar( HEXITS[((0xF0 & in)>>4)] );
    txChar( HEXITS[(0x0F & in)] );
}

void txVal(char *s, unsigned char val)
{
    txStr(s);
    txHex(val);
    txChar('\r');
}

void txArray(unsigned char *array, unsigned char length)
{
    unsigned char i;
    for (i=0; i < length; i++)
    {
        txHex(array[i]);
        txChar(' ');
    }
}

// RX interrupt on any character
void __attribute__ ((interrupt,no_auto_psv)) _U1RXInterrupt(void)
{
    IFS0bits.U1RXIF = 0;            // clear interrupt flag

    while (U1STAbits.URXDA) // while the RX buffer has stuff in it
    {
       UART_RX_buffer[UART_RX_buffer_size++] = U1RXREG; // put it in our own buffer
    }

    U1RX_flag = 1;      // set a flag for the main loop to know we have new data

    if (U1STAbits.OERR) U1STAbits.OERR = 0; // clear buffer errors
}

// TX interrupt when last character is shifted out of TSR
void __attribute__ ((interrupt,no_auto_psv)) _U1TXInterrupt(void)
{
    IFS0bits.U1TXIF = 0;
    U1TX_flag = 1;

    if (tx_buffer_len)  // do we have more characters to send?
    {
        U1TXREG = UART_TX_buffer[tx_buffer_pos++];   // send the next character
        if (tx_buffer_pos >= tx_buffer_len) // have we sent all the characters in our buffer?
        {
            tx_buffer_pos = tx_buffer_len = tx_fifo = 0;    // yep, we're done, clear everything
        }
    }  else if (tx_fifo) {  // no more characters to send, so we have an opening in the FIFO
        tx_fifo--;
    }
}

